# SPDX-FileCopyrightText: Copyright (c) 2023, Dask Developers, Yuvi Panda, Anaconda Inc, NVIDIA
# SPDX-License-Identifier: BSD 3-Clause License
from kr8s import HTTPClient, KubeConfig
from kr8s.objects import Pod
from kr8s.query import Query


async def test_pod_query():
    api = HTTPClient(KubeConfig.from_env())
    async for pod in Query(api, Pod, namespace="kube-system"):
        assert isinstance(pod, Pod)


async def test_pod_objects():
    api = HTTPClient(KubeConfig.from_env())
    async for pod in Pod.objects(api).filter(namespace="kube-system"):
        assert isinstance(pod, Pod)



"""An asyncio shim for pykube-ng."""
# SPDX-FileCopyrightText: Copyright (c) 2023, Dask Developers, Yuvi Panda, Anaconda Inc, NVIDIA
# SPDX-License-Identifier: BSD 3-Clause License
from pykube.query import Query as _Query
from pykube.query import Table, now
from pykube.query import WatchQuery as _WatchQuery

from kr8s.mixins import AsyncMixin


class Query(_Query, AsyncMixin):
    async def execute(self, **kwargs):
        return await self._sync(super().execute, **kwargs)

    async def iterator(self):
        response = await self.execute()
        for obj in response.json().get("items") or []:
            yield self.api_obj_class(self.api, obj)

    def __aiter__(self):
        return self.iterator()

    async def get_by_name(self, name: str):
        return await self._sync(super().get_by_name, name=name)

    async def get(self, *args, **kwargs):
        return await self._sync(super().get, *args, **kwargs)

    async def get_or_none(self, *args, **kwargs):
        return await self._sync(super().get_or_none, *args, **kwargs)

    async def as_table(self) -> Table:
        response = await self.execute(
            headers={"Accept": "application/json;as=Table;v=v1beta1;g=meta.k8s.io"}
        )
        return Table(self.api_obj_class, response.json())

    def watch(self, since=None, *, params=None):
        query = self._clone(WatchQuery)
        query.params = params
        if since is now:
            raise ValueError("now is not a supported since value in async version")
        elif since is not None:
            query.resource_version = since
        return query

    @property
    def query_cache(self):
        raise NotImplementedError(
            "Properties cannot make async HTTP requests to populate the cache. "
            "Also the shim currently does not implement a cache at all."
        )

    @property
    def response(self):
        raise NotImplementedError(
            "Properties cannot make HTTP requests."
            "Use ``response = (await Query.execute()).json()`` instead."
        )

    def __len__(self):
        raise TypeError(
            "Cannot call len directly on async objects. "
            "Instead you can use ``len([_ async for _ in Query(...)])``."
        )


class WatchQuery(_WatchQuery, AsyncMixin):
    async def object_stream(self):
        f = super().object_stream
        object_stream = await self._sync(lambda: iter(f()))
        while True:
            try:
                yield await self._sync(next, object_stream)
            except StopIteration:
                break

    def __aiter__(self):
        return self.object_stream()



"""An asyncio shim for pykube-ng."""
# SPDX-FileCopyrightText: Copyright (c) 2023, Dask Developers, Yuvi Panda, Anaconda Inc, NVIDIA
# SPDX-License-Identifier: BSD 3-Clause License
from pykube import HTTPClient, KubeConfig, all  # noqa

__version__ = "0.1.0-a.0"



# SPDX-FileCopyrightText: Copyright (c) 2023, Dask Developers, Yuvi Panda, Anaconda Inc, NVIDIA
# SPDX-License-Identifier: BSD 3-Clause License
import asyncio
import functools

from pykube.mixins import ScalableMixin as _ScalableMixin


class AsyncMixin:
    async def _sync(self, func, *args, **kwargs):
        return await asyncio.get_event_loop().run_in_executor(
            None, functools.partial(func, *args, **kwargs)
        )


class AsyncScalableMixin(_ScalableMixin):
    async def scale(self, replicas=None):
        count = self.scalable if replicas is None else replicas
        await self.exists(ensure=True)
        if self.scalable != count:
            self.scalable = count
            await self.update()
            while True:
                await self.reload()
                if self.scalable == count:
                    break
                await asyncio.sleep(1)



# SPDX-FileCopyrightText: Copyright (c) 2023, Dask Developers, Yuvi Panda, Anaconda Inc, NVIDIA
# SPDX-License-Identifier: BSD 3-Clause License
from typing import Optional

from pykube.http import HTTPClient
from pykube.objects import APIObject as _APIObject
from pykube.objects import ClusterRole as _ClusterRole
from pykube.objects import ClusterRoleBinding as _ClusterRoleBinding
from pykube.objects import ConfigMap as _ConfigMap
from pykube.objects import CronJob as _CronJob
from pykube.objects import CustomResourceDefinition as _CustomResourceDefinition
from pykube.objects import DaemonSet as _DaemonSet
from pykube.objects import Deployment as _Deployment
from pykube.objects import Endpoint as _Endpoint
from pykube.objects import Event as _Event
from pykube.objects import HorizontalPodAutoscaler as _HorizontalPodAutoscaler
from pykube.objects import Ingress as _Ingress
from pykube.objects import Job as _Job
from pykube.objects import LimitRange as _LimitRange
from pykube.objects import Namespace as _Namespace
from pykube.objects import NamespacedAPIObject as _NamespacedAPIObject
from pykube.objects import Node as _Node
from pykube.objects import ObjectManager
from pykube.objects import PersistentVolume as _PersistentVolume
from pykube.objects import PersistentVolumeClaim as _PersistentVolumeClaim
from pykube.objects import Pod as _Pod
from pykube.objects import PodDisruptionBudget as _PodDisruptionBudget
from pykube.objects import PodSecurityPolicy as _PodSecurityPolicy
from pykube.objects import ReplicaSet as _ReplicaSet
from pykube.objects import ReplicationController as _ReplicationController
from pykube.objects import ResourceQuota as _ResourceQuota
from pykube.objects import Role as _Role
from pykube.objects import RoleBinding as _RoleBinding
from pykube.objects import Secret as _Secret
from pykube.objects import Service as _Service
from pykube.objects import ServiceAccount as _ServiceAccount
from pykube.objects import StatefulSet as _StatefulSet
from requests import Response

from kr8s.mixins import AsyncMixin, AsyncScalableMixin
from kr8s.query import Query


class AsyncObjectManager(ObjectManager):
    def __call__(self, api: HTTPClient, namespace: str = None):
        query = super().__call__(api=api, namespace=namespace)
        return query._clone(Query)


class AsyncObjectMixin(AsyncMixin):
    objects = AsyncObjectManager()

    async def exists(self, ensure=False):
        return await self._sync(super().exists, ensure=ensure)

    async def create(self):
        return await self._sync(super().create)

    async def reload(self):
        return await self._sync(super().reload)

    def watch(self):
        return super().watch()

    async def patch(self, strategic_merge_patch, *, subresource=None):
        return await self._sync(
            super().patch, strategic_merge_patch, subresource=subresource
        )

    async def update(self, is_strategic=True, *, subresource=None):
        return await self._sync(super().update, is_strategic, subresource=subresource)

    async def delete(self, propagation_policy: str = None):
        return await self._sync(super().delete, propagation_policy=propagation_policy)

    exists.__doc__ = _APIObject.exists.__doc__
    create.__doc__ = _APIObject.create.__doc__
    reload.__doc__ = _APIObject.reload.__doc__
    watch.__doc__ = _APIObject.watch.__doc__
    patch.__doc__ = _APIObject.patch.__doc__
    update.__doc__ = _APIObject.update.__doc__
    delete.__doc__ = _APIObject.delete.__doc__


class APIObject(AsyncObjectMixin, _APIObject):
    """APIObject."""


class NamespacedAPIObject(AsyncObjectMixin, _NamespacedAPIObject):
    """APIObject."""


class ConfigMap(AsyncObjectMixin, _ConfigMap):
    """ConfigMap."""


class CronJob(AsyncObjectMixin, _CronJob):
    """CronJob."""


class DaemonSet(AsyncObjectMixin, _DaemonSet):
    """DaemonSet."""


class Deployment(AsyncScalableMixin, AsyncObjectMixin, _Deployment):
    """Deployment."""

    async def rollout_undo(self, target_revision=None):
        return await self._sync(super().rollout_undo, target_revision)

    rollout_undo.__doc__ = _Deployment.rollout_undo.__doc__


class Endpoint(AsyncObjectMixin, _Endpoint):
    """Endpoint."""


class Event(AsyncObjectMixin, _Event):
    """Event."""


class LimitRange(AsyncObjectMixin, _LimitRange):
    """LimitRange."""


class ResourceQuota(AsyncObjectMixin, _ResourceQuota):
    """ResourceQuota."""


class ServiceAccount(AsyncObjectMixin, _ServiceAccount):
    """ServiceAccount."""


class Ingress(AsyncObjectMixin, _Ingress):
    """Ingress."""


class Job(AsyncScalableMixin, AsyncObjectMixin, _Job):
    """Job."""


class Namespace(AsyncObjectMixin, _Namespace):
    """Namespace."""


class Node(AsyncObjectMixin, _Node):
    """Node."""


class Pod(AsyncObjectMixin, _Pod):
    """Pod"""

    async def logs(self, *args, **kwargs):
        return await self._sync(super().logs, *args, **kwargs)

    logs.__doc__ = _Pod.logs.__doc__


class ReplicationController(
    AsyncScalableMixin, AsyncObjectMixin, _ReplicationController
):
    """ReplicationController."""


class ReplicaSet(AsyncScalableMixin, AsyncObjectMixin, _ReplicaSet):
    """ReplicaSet."""


class Secret(AsyncObjectMixin, _Secret):
    """Secret."""


class Service(AsyncObjectMixin, _Service):
    """Service."""

    async def proxy_http_request(self, *args, **kwargs):
        return await self._sync(super().proxy_http_request, *args, **kwargs)

    async def proxy_http_get(
        self, path: str, port: Optional[int] = None, **kwargs
    ) -> Response:
        return await self.proxy_http_request("GET", path, port, **kwargs)

    async def proxy_http_post(
        self, path: str, port: Optional[int] = None, **kwargs
    ) -> Response:
        return await self.proxy_http_request("POST", path, port, **kwargs)

    async def proxy_http_put(
        self, path: str, port: Optional[int] = None, **kwargs
    ) -> Response:
        return await self.proxy_http_request("PUT", path, port, **kwargs)

    async def proxy_http_delete(
        self, path: str, port: Optional[int] = None, **kwargs
    ) -> Response:
        return await self.proxy_http_request("DELETE", path, port, **kwargs)

    proxy_http_request.__doc__ = _Service.proxy_http_request.__doc__
    proxy_http_get.__doc__ = _Service.proxy_http_get.__doc__
    proxy_http_post.__doc__ = _Service.proxy_http_post.__doc__
    proxy_http_put.__doc__ = _Service.proxy_http_put.__doc__
    proxy_http_delete.__doc__ = _Service.proxy_http_delete.__doc__


class PersistentVolume(AsyncObjectMixin, _PersistentVolume):
    """PersistentVolume."""


class PersistentVolumeClaim(AsyncObjectMixin, _PersistentVolumeClaim):
    """PersistentVolumeClaim."""


class HorizontalPodAutoscaler(AsyncObjectMixin, _HorizontalPodAutoscaler):
    """HorizontalPodAutoscaler."""


class StatefulSet(AsyncScalableMixin, AsyncObjectMixin, _StatefulSet):
    """StatefulSet."""


class Role(AsyncObjectMixin, _Role):
    """Role."""


class RoleBinding(AsyncObjectMixin, _RoleBinding):
    """RoleBinding."""


class ClusterRole(AsyncObjectMixin, _ClusterRole):
    """ClusterRole."""


class ClusterRoleBinding(AsyncObjectMixin, _ClusterRoleBinding):
    """ClusterRoleBinding."""


class PodSecurityPolicy(AsyncObjectMixin, _PodSecurityPolicy):
    """PodSecurityPolicy."""


class PodDisruptionBudget(AsyncObjectMixin, _PodDisruptionBudget):
    """PodDisruptionBudget."""


class CustomResourceDefinition(AsyncObjectMixin, _CustomResourceDefinition):
    """CustomResourceDefinition."""



# SPDX-FileCopyrightText: Copyright (c) 2023, Dask Developers, Yuvi Panda, Anaconda Inc, NVIDIA
# SPDX-License-Identifier: BSD 3-Clause License
import os

import pytest
from pytest_kind.cluster import KindCluster


@pytest.fixture(scope="session", autouse=True)
def k8s_cluster(request):
    image = None
    if version := os.environ.get("KUBERNETES_VERSION"):
        image = f"kindest/node:v{version}"

    kind_cluster = KindCluster(
        name="pytest-kind",
        image=image,
    )
    kind_cluster.create()
    os.environ["KUBECONFIG"] = str(kind_cluster.kubeconfig_path)
    yield kind_cluster
    del os.environ["KUBECONFIG"]
    if not request.config.getoption("keep_cluster"):
        kind_cluster.delete()



# SPDX-FileCopyrightText: Copyright (c) 2023, Dask Developers, Yuvi Panda, Anaconda Inc, NVIDIA
# SPDX-License-Identifier: BSD 3-Clause License
import asyncio
import uuid

from kr8s import HTTPClient, KubeConfig
from kr8s.objects import Pod


async def test_pod_create_and_delete():
    api = HTTPClient(KubeConfig.from_env())
    name = "test-" + uuid.uuid4().hex[:10]
    pod = Pod(
        api,
        {
            "apiVersion": "v1",
            "kind": "Pod",
            "metadata": {"name": name},
            "spec": {
                "containers": [
                    {"name": "pause", "image": "gcr.io/google_containers/pause"}
                ]
            },
        },
    )

    await pod.create()
    assert await pod.exists()
    await pod.delete()
    while await pod.exists():
        await asyncio.sleep(0.1)
    assert not await pod.exists()
